﻿using System;
using System.Collections.Generic;
using System.Text;

namespace CVUT.Vorel.elasticity
{
  class SC
  {

    public static Tensor SC_routine(Matrix A_matrix, Inclusion A_inclusion, double Scale)
    {
      if (Scale <= 0)
      {
        throw new Exception("Elasticity: SC: SC_routine: wrong Scale!");
      }
      Tensor Tr = new Tensor(6, 6);
      Tensor L = new Tensor(6, 6);
      Tensor L0 = new Tensor(A_matrix.StiffnessMatrix);
      Tensor Li = new Tensor(A_inclusion.StiffnessMatrix); 
      Tensor ai = new Tensor(A_inclusion.SemiAxes);
      Matrix dum_matrix = new Matrix(A_matrix);
      double ci = (A_inclusion.VolFrac / Scale);
      double err = 1.0, tol0 = 1.0e-6, tol;
      int maxIter = 300, nIter = 0;

      Li = A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStressInverse() * Li *
           A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrain();

      tol   = tol0 * Math.Min(Tensor.MinAbsNonZeroElement(dum_matrix.StiffnessMatrix), Tensor.MaxAbsElement(Li));
      err   = Tensor.MaxAbsElement(dum_matrix.StiffnessMatrix);
      nIter = 0;
      try
      {
        while ((err > tol && nIter < maxIter) || nIter == 0)
        {
          nIter++;

          Tr = ConcentrationFactor(A_inclusion, dum_matrix);
          L = L0 + ci * (Li - L0) * Tr;
          // symmetrize matrix
          L = 0.5 * (L + L.Transposition());

          if (nIter == 1)
          {
            tol = Math.Min(tol, tol0 * Tensor.MaxElement(L));
          }
          err = Tensor.MaxAbsElement(L - dum_matrix.StiffnessMatrix);
          dum_matrix.StiffnessMatrix = L;
        }
        if (nIter >= maxIter)
        {
          throw new Exception("Elasticity: SC: SC_routine: Maximum number (" + Convert.ToString(maxIter, Constants.Separators.FormatProvider) +
                              " of iterations was exceeded!");
        }
      }
      catch
      {
        L0 = A_inclusion.StiffnessMatrix;
        Li = A_matrix.StiffnessMatrix;

        // rotation adjustment
        L0 = A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStressInverse() * L0 *
             A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrain();   // return to global coordinate system

        Li = A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStress() * Li *
             A_inclusion.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrainInverse(); // rotate to local

        Inclusion dumInclusion = new Inclusion(A_inclusion);
        dumInclusion.StiffnessMatrix = Li;

        dum_matrix = new Matrix(A_matrix);
        dum_matrix.StiffnessMatrix = L0;

        ci = 1 - ci;

        err = Tensor.MaxAbsElement(dum_matrix.StiffnessMatrix);
        nIter = 0;
        while ((err > tol && nIter < maxIter) || nIter == 0)
        {
          nIter++;

          Tr = ConcentrationFactor(dumInclusion, dum_matrix);
          L = L0 + ci * (Li - L0) * Tr;
          // symmetrize matrix
          L = 0.5 * (L + L.Transposition());

          err = Tensor.MaxAbsElement(L - dum_matrix.StiffnessMatrix);
          dum_matrix.StiffnessMatrix = L;
        }
        if (nIter >= maxIter)
        {
          throw new Exception("Elasticity: SC: SC_routine: Maximum number (" + Convert.ToString(maxIter, Constants.Separators.FormatProvider) +
                              ") of iterations was exceeded!");
        }
      }

      return (L);
    }
    public static Tensor SC_routine(Matrix A_matrix, List<Inclusion> A_list, double Scale)
    {
      Tensor L = new Tensor(6, 6);

      if (A_list.Count == 0)
      {
        L = A_matrix.StiffnessMatrix;
      }
      else if (A_list.Count == 1)
      {
        L = SC_routine(A_matrix, A_list[0], Scale);
      }
      else if (A_list.Count > 1)
      {

        if (Scale <= 0)
        {
          throw new Exception("Elasticity: SC: SC_routine: wrong Scale!");
        }

        Tensor Tr = new Tensor(6, 6);
        Matrix dumMatrix = new Matrix(A_matrix);
        Tensor L0 = new Tensor(A_matrix.StiffnessMatrix);
        Tensor Li = new Tensor(6, 6);
        Tensor f = new Tensor(6, 6);
        Tensor U = new Tensor(6, 6);
        Tensor I = new Tensor(6, 6);
        double ci, ci_sum = 0.0;
        double err = 1.0, tol0 = 1.0e-6, tol, dumMax, dum;
        int maxIter = 10000, nIter = 0;

        dumMax = -1.0;
        for (int n = 0; n < A_list.Count; n++)
        {
          dum = Tensor.MaxAbsElement(A_list[n].StiffnessMatrix);
          if( dum > dumMax ) 
          {
            dumMax = dum;
          }
        }

        if (dumMax > 0)
        {
          tol = tol0 * Math.Min(Tensor.MinAbsNonZeroElement(dumMatrix.StiffnessMatrix), dumMax);
        } 
        else
        {
          tol = tol0 * Tensor.MinAbsNonZeroElement(dumMatrix.StiffnessMatrix);
        }
        err = Tensor.MaxAbsElement(dumMatrix.StiffnessMatrix);
        nIter = 0;

        while ((err > tol && nIter < maxIter) || nIter == 0)
        {
          nIter++;
          ci_sum = 0.0;
          U.Clear();
          for (int n = 0; n < A_list.Count; n++)
          {
            ci = (A_list[n].VolFrac / Scale);
            Li = A_list[n].StiffnessMatrix;
            Li = A_list[n].EulerAngles.TransformationEulerAngles_zyz().TransformationForStressInverse() * Li *
                 A_list[n].EulerAngles.TransformationEulerAngles_zyz().TransformationForStrain();

            Tr = ConcentrationFactor(A_list[n], dumMatrix);

            f = Li - L0;
            U += Tensor.ConstantTensorProduct(ci, Tensor.TensorTensorProduct(f, Tr));

            ci_sum += ci;
          }

          if (ci_sum > 1)
          {
            throw new Exception("Elasticity: SC: SC_routine: volumu of inclusions is > 1!");
          }

          L = L0 + U;
          // symmetrize matrix
          L = 0.5 * (L + L.Transposition());

          err = Tensor.MaxAbsElement(L - dumMatrix.StiffnessMatrix);
          dumMatrix.StiffnessMatrix = L;

        }

        if (nIter >= maxIter)
        {
          throw new Exception("Elasticity: SC: SC_routine: Maximum number (" + Convert.ToString(maxIter, Constants.Separators.FormatProvider) +
                              ") of iterations was exceeded! - try to swap the phases");
        }

      }

      return (L);
    }

    private static Tensor ConcentrationFactor(Inclusion incl, Matrix matr)
    {
      Tensor Tr = new Tensor(6, 6);
      Tensor L0 = new Tensor(matr.StiffnessMatrix);
      Tensor S = new Tensor(6, 6);
      Tensor Li = new Tensor(incl.StiffnessMatrix);


      if (matr.MaterialKind == Constants.EMaterialKind.Isotropic)
      {
        Tensor dummy = new Tensor(2);
        dummy = matr.ConstantsFromStiffnessMatrix();

        S = Inclusion.GetEshelbyTensorIso(incl.SemiAxes, dummy[1]);
      }
      else if (matr.MaterialKind != Constants.EMaterialKind.FullyAnisotropic)
      {
        L0 = incl.EulerAngles.TransformationEulerAngles_zyz().TransformationForStress() * L0 *
             incl.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrainInverse();

        S = Inclusion.GetEshelbyTensorAniso(incl.SemiAxes, L0.Conversion2ndTo4thOrderStiffness(), 15, 40);
      }
      else
      {
        throw new Exception("Elasticity: SC: ConcentrationFactor: Fully anisotropic stiffness matrix!!!");
      }

      Tr = (Tensor.GetIdentityTensor(6) + S * L0.InverseMatrix() * (Li - L0));
      Tr = Tr.InverseMatrix();

      Tr = incl.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrainInverse() * Tr * 
           incl.EulerAngles.TransformationEulerAngles_zyz().TransformationForStrain();   // return to global coordinate system

      return Tr;
    }
    private static Tensor ConcentrationFactor(Tensor semiAxes, Tensor Li, Tensor L0)
    {
      if(semiAxes.which_order!=1||semiAxes.how_many_lines!=3)
      {
        throw new Exception("Elasticity: SC: ConcentrationFactor: wrong tensor!!!");
      }
      else if (L0.which_order != 2 || L0.how_many_lines != 6 || L0.how_many_columns != 6)
      {
        throw new Exception("Elasticity: SC: ConcentrationFactor: wrong tensor!!!");
      }
      else if (Li.which_order != 2 || Li.how_many_lines != 6 || Li.how_many_columns != 6)
      {
        throw new Exception("Elasticity: SC: ConcentrationFactor: wrong tensor!!!");
      }

      Matrix dummyMatrix = new Matrix(L0);
      Tensor Tr = new Tensor(6, 6);
      Tensor S = new Tensor(6, 6);

      if (dummyMatrix.MaterialKind == Constants.EMaterialKind.Isotropic)
      {
        Tensor dummy = new Tensor(2);
        dummy = dummyMatrix.ConstantsFromStiffnessMatrix();

        S = Inclusion.GetEshelbyTensorIso(semiAxes, dummy[1]);
      }
      else if (dummyMatrix.MaterialKind != Constants.EMaterialKind.FullyAnisotropic)
      {
        S = Inclusion.GetEshelbyTensorAniso(semiAxes, L0.Conversion2ndTo4thOrderStiffness(), 15, 40);
      }
      else
      {
        throw new Exception("Elasticity: SC: ConcentrationFactor: Fully anisotropic stiffness matrix!!!");
      }

      Tr = (Tensor.GetIdentityTensor(6) + S * L0.InverseMatrix() * (Li - L0));
      Tr = Tr.InverseMatrix();

      return Tr;
    }
    private static Tensor OrientationAverage(Tensor A, double Theta, string type)  //It is assumed that \beta=0 and \phi=0
    {
      if (A.which_order != 2 || A.how_many_lines != 6 || A.how_many_columns != 6)
      {
        throw new Exception("Elasticity: SC: OrientationAverage: wrong tensor!");
      }

      // sin(x) = 2;
      // cos(x) = 3;
      // (cos(x))^2-(sin(x))^2 = 5;

      Tensor P1 = new Tensor(6, 6);
      int i, j, k, l;
      double T12;

      double int_s = 0.0;
      double int_ss = Theta - 0.5 * Math.Sin(2 * Theta);
      double int_sss = 0.0;
      double int_ssss = 3 * Theta / 4 - Math.Sin(2 * Theta) / 2 + Math.Sin(4 * Theta) / 16;

      double int_c = 2 * Math.Sin(Theta);
      double int_cc = Theta + 0.5 * Math.Sin(2 * Theta);
      double int_ccc = 2 * Math.Sin(Theta) - 2 * Math.Pow(Math.Sin(Theta), 3) / 3;
      double int_cccc = 3 * Theta / 4 + Math.Sin(2 * Theta) / 2 + Math.Sin(4 * Theta) / 16;

      double int_sc = 0.0;
      double int_scc = 0.0;
      double int_sccc = 0.0;

      double int_ssc = 2 * Math.Pow(Math.Sin(Theta), 3) / 3;
      double int_sscc = Theta / 4 - Math.Sin(4 * Theta) / 16;

      double int_sssc = 0;


      int[,] T = new int[6, 6];
      int[,] Tt = new int[6, 6];

      //         cos(theta)^2, 0,           sin(theta)^2,           0,  -2*cos(theta)*sin(theta),          0;
      //                    0, 1,                      0,           0,                         0,          0;
      //         sin(theta)^2, 0,           cos(theta)^2,           0,   2*cos(theta)*sin(theta),          0;
      //                    0, 0,                      0,  cos(theta),                         0, sin(theta);
      //cos(theta)*sin(theta), 0, -cos(theta)*sin(theta),           0, cos(theta)^2-sin(theta)^2,          0;
      //                    0, 0,                      0, -sin(theta),                         0, cos(theta)
    
        T[0, 0] = 9; T[0, 1] = 0; T[0, 2] = 4; T[0, 3] = 0; T[0, 4] = -6; T[0, 5] = 0;
        T[1, 0] = 0; T[1, 1] = 1; T[1, 2] = 0; T[1, 3] = 0; T[1, 4] = 0; T[1, 5] = 0;
        T[2, 0] = 4; T[2, 1] = 0; T[2, 2] = 9; T[2, 3] = 0; T[2, 4] = 6; T[2, 5] = 0;
        T[3, 0] = 0; T[3, 1] = 0; T[3, 2] = 0; T[3, 3] = 3; T[3, 4] = 0; T[3, 5] = 2;
        T[4, 0] = 6; T[4, 1] = 0; T[4, 2] = -6; T[4, 3] = 0; T[4, 4] = 9 - 4; T[4, 5] = 0;
        T[5, 0] = 0; T[5, 1] = 0; T[5, 2] = 0; T[5, 3] = -2; T[5, 4] = 0; T[5, 5] = 3;

        Tt[0, 0] = 9; Tt[0, 1] = 0; Tt[0, 2] = 4; Tt[0, 3] = 0; Tt[0, 4] = 6; Tt[0, 5] = 0;
        Tt[1, 0] = 0; Tt[1, 1] = 1; Tt[1, 2] = 0; Tt[1, 3] = 0; Tt[1, 4] = 0; Tt[1, 5] = 0;
        Tt[2, 0] = 4; Tt[2, 1] = 0; Tt[2, 2] = 9; Tt[2, 3] = 0; Tt[2, 4] = -6; Tt[2, 5] = 0;
        Tt[3, 0] = 0; Tt[3, 1] = 0; Tt[3, 2] = 0; Tt[3, 3] = 3; Tt[3, 4] = 0; Tt[3, 5] = -2;
        Tt[4, 0] = -6; Tt[4, 1] = 0; Tt[4, 2] = 6; Tt[4, 3] = 0; Tt[4, 4] = 9 - 4; Tt[4, 5] = 0;
        Tt[5, 0] = 0; Tt[5, 1] = 0; Tt[5, 2] = 0; Tt[5, 3] = 2; Tt[5, 4] = 0; Tt[5, 5] = 3;


 //           cos(theta)^2, 0,             sin(theta)^2,           0,    -cos(theta)*sin(theta),          0;
 //                      0, 1,                        0,           0,                         0,          0;
 //           sin(theta)^2, 0,             cos(theta)^2,           0,     cos(theta)*sin(theta),          0;
 //                      0, 0,                        0,  cos(theta),                         0, sin(theta);
 //2*cos(theta)*sin(theta), 0, -2*cos(theta)*sin(theta),           0, cos(theta)^2-sin(theta)^2,          0;
 //                      0, 0,                        0, -sin(theta),                         0, cos(theta);



      for (i = 0; i < 6; i++)
      {
        for (j = 0; j < 6; j++)
        {
          for (k = 0; k < 6; k++)
          {
            for (l = 0; l < 6; l++)
            {
              switch (Tt[i, k] * T[l, j])
              {
                case 0:
                  T12 = 0.0;
                  break;
                case 1:
                  T12 = 2.0 * Theta;
                  break;
                case 2:
                  T12 = int_s;
                  break;
                case 4:
                  T12 = int_ss;
                  break;
                case 8:
                  T12 = int_sss;
                  break;
                case 16:
                  T12 = int_ssss;
                  break;

                case 3:
                  T12 = int_c;
                  break;
                case 9:
                  T12 = int_cc;
                  break;
                case 27:
                  T12 = int_ccc;
                  break;
                case 81:
                  T12 = int_cccc;
                  break;

                case 6:
                  T12 = int_sc;
                  break;
                case 18:
                  T12 = int_scc;
                  break;
                case 54:
                  T12 = int_sccc;
                  break;

                case 12:
                  T12 = int_ssc;
                  break;
                case 36:
                  T12 = int_sscc;
                  break;

                case 24:
                  T12 = int_sssc;
                  break;

                case 10:
                  T12 = int_scc - int_sss;
                  break;
                case 15:
                  T12 = int_ccc - int_ssc;
                  break;
                case 20:
                  T12 = int_sscc - int_ssss;
                  break;
                case 45:
                  T12 = int_cccc - int_sscc;
                  break;
                case 30:
                  T12 = int_sccc - int_sssc;
                  break;
                case 5:
                  T12 = int_cc - int_ss;
                  break; 
                case 25:
                  T12 = int_cccc - 2 * int_sscc + int_ssss;
                  break;


                case -1:
                  T12 = -2.0 * Theta;
                  break;
                case -2:
                  T12 = -int_s;
                  break;
                case -4:
                  T12 = -int_ss;
                  break;
                case -8:
                  T12 = -int_sss;
                  break;
                case -16:
                  T12 = -int_ssss;
                  break;

                case -3:
                  T12 = -int_c;
                  break;
                case -9:
                  T12 = -int_cc;
                  break;
                case -27:
                  T12 = -int_ccc;
                  break;
                case -81:
                  T12 = -int_cccc;
                  break;

                case -6:
                  T12 = -int_sc;
                  break;
                case -18:
                  T12 = -int_scc;
                  break;
                case -54:
                  T12 = -int_sccc;
                  break;

                case -12:
                  T12 = -int_ssc;
                  break;
                case -36:
                  T12 = -int_sscc;
                  break;

                case -24:
                  T12 = -int_sssc;
                  break;

                case -10:
                  T12 = -int_scc + int_sss;
                  break;
                case -15:
                  T12 = -int_ccc + int_ssc;
                  break;
                case -20:
                  T12 = -int_sscc + int_ssss;
                  break;
                case -45:
                  T12 = -int_cccc + int_sscc;
                  break;
                case -30:
                  T12 = -int_sccc + int_sssc;
                  break;
                case -5:
                  T12 = -int_cc + int_ss;
                  break; 

                default:
                  throw new Exception("Elasticity: SC: OrientationAverage: wrong case!");
              }

              // to grasp some elements of transformation matrix, where the number two is
              switch (type)
              {
                case "Li":
                  if ((i == 0 || i == 2) && k == 4)
                  {
                    T12 = 2 * T12;
                  }
                  if ((j == 0 || j == 2) && l == 4)
                  {
                    T12 = 2 * T12;
                  }
                  break;

                case "Tr":
                  if ((k == 0 || k == 2) && i == 4)
                  {
                    T12 = 2 * T12;
                  }
                  if ((j == 0 || j == 2) && l == 4)
                  {
                    T12 = 2 * T12;
                  }
                  break;

                default:
                  throw new Exception("Elasticity: SC: OrientationAverage: Unsupported case!!!");
              }
              P1[i, j] += 1 / (2 * Theta) * T12 * A[k, l];
            }
          }
        }
      }
      return (P1);
    }
    private static Tensor OrientationAverage(Tensor A, DataXY histogram, string type)
    {
      if (A.which_order != 2 || A.how_many_lines != 6 || A.how_many_columns != 6)
      {
        throw new Exception("Elasticity: SC: OrientationAverage: wrong tensor!");
      }

      Tensor P1 = new Tensor(6, 6);
      Tensor T1 = new Tensor(6, 6);
      Tensor angles = new Tensor(3); //for euler angles only second angle is not equal 0 (rotation around second axes

      for (int i = 0; i < histogram.Count; i++)
      {
        angles[1] = histogram.x[i];

        switch (type)
        {
          case "Li":
            P1 = P1 + histogram.y[i] *
                 angles.TransformationEulerAngles_zyz().TransformationForStressInverse() * A * 
                 angles.TransformationEulerAngles_zyz().TransformationForStrain();
            break;

          case "Tr":
            P1 = P1 + histogram.y[i] *
                 angles.TransformationEulerAngles_zyz().TransformationForStrainInverse() * A * 
                 angles.TransformationEulerAngles_zyz().TransformationForStrain();
            break;

          default:
            throw new Exception("Elasticity: SC: OrientationAverage: Unsupported case!!!");
        }
      }

      return (P1);
    }

  }
}
