﻿using System;
using System.Collections.Generic;
using System.Xml;
using System.Text;

namespace CVUT.Vorel.elasticity
{
  public class Inclusion : Constituent
  {
    private Tensor _a;																							// semi-axes
    private double _VolumeFrac;																			// volume fracture of the whole structure
    private Tensor _EulerAngles;																		// Euler angles-phi, theta, beta
    private bool   _RandomOrientation;			    										// defines if the random orientation will be used
    public string name = "";

    /////////////////////////////////////////////////////////////////////////////////////////////
    public Inclusion()
      : base()
    {
      _a = new Tensor(3);
      _a[0] = _a[1] = _a[2] = 1;
      _EulerAngles = new Tensor(3);
      
      _VolumeFrac = 0.0;
      _RandomOrientation = false;

    }
    public Inclusion(double a0, double a1, double a2)
      : base()
    {
      _a = new Tensor(3);
      _EulerAngles = new Tensor(3);

      _a[0] = a0;
      _a[1] = a1;
      _a[2] = a2;

      _VolumeFrac = 0.0;
      _RandomOrientation = false;
    }
    public Inclusion(double a0, double a1, double a2, double E, double nu, double VolumeFr)
      : base(E, nu)
    {
      _a = new Tensor(3);
      _EulerAngles = new Tensor(3);

      _a[0] = a0;
      _a[1] = a1;
      _a[2] = a2;

      _VolumeFrac = VolumeFr;
      _RandomOrientation = false;
    }
    public Inclusion(double a0, double a1, double a2, Tensor StiffMatrix, double VolumeFr)
      : base(StiffMatrix)
    {
      _a = new Tensor(3);
      _EulerAngles = new Tensor(3);

      _a[0] = a0;
      _a[1] = a1;
      _a[2] = a2;

      _VolumeFrac = VolumeFr;
      _RandomOrientation = false;
    }
    public Inclusion(Inclusion oa)
      : base(oa)
    {
      _a = new Tensor(oa._a);
      _EulerAngles = new Tensor(oa._EulerAngles);

      _VolumeFrac = oa._VolumeFrac;
      _RandomOrientation = oa._RandomOrientation;
    }
    public Inclusion(Tensor a, Tensor StiffMatrix, double VolumeFr, Tensor EulerAngles)
      : base(StiffMatrix)
    {
      _a = new Tensor(a);
      _EulerAngles = new Tensor(EulerAngles);

      _VolumeFrac = VolumeFr;
      _RandomOrientation = false;
    }
    /////////////////////////////////////////////////////////////////////////////////////////////
    public Tensor SemiAxes
    {
      get
      {
        return (_a);
      }
    }
    public Tensor EulerAngles
    {
      get
      {
        return (_EulerAngles);
      }
    }
    public double VolFrac
    {
      get
      {
        return (_VolumeFrac);
      }
      set
      {
        _VolumeFrac = value;
      }
    }
    public bool RandomOrientation
    {
      get
      {
        return (_RandomOrientation);
      }
      set
      {
        _RandomOrientation = value;
      }
    }
    public Tensor ShapeTensor
    {
      get
      {
        Tensor L = new Tensor(3, 3);

        L[0, 0] = 1 / (_a[0] * _a[0]);
        L[1, 1] = 1 / (_a[1] * _a[1]);
        L[2, 2] = 1 / (_a[2] * _a[2]);

        return (L);
      }
    }
    public Constants.EInclusionKind InclusionKind
    {
      get
      {
        return this.GetShapeKind();
      }
    }
    private Constants.EInclusionKind GetShapeKind()
    {
      Tensor a = new Tensor(this._a);
      int seq;

      a = a.Sort(out seq);

      if (Math.Abs(a[0] - a[1]) < Constants.ZERO && Math.Abs(a[0] - a[2]) < Constants.ZERO)
      {
        return (Constants.EInclusionKind.Sphere);
      }
      else if (a[0] == Constants.INFINITY)
      {
        if (Math.Abs(a[1] - a[2]) < Constants.ZERO)
        {
          return (Constants.EInclusionKind.Cylinder);
        }
        else
        {
          return (Constants.EInclusionKind.EllipticCylinder);
        }
      }
      else if (Math.Abs(a[0] - a[1]) < Constants.ZERO && Functions.IsMuchMore(a[0], a[2]))
      {
        return (Constants.EInclusionKind.PennyShape);
      }
      else
      {
        return (Constants.EInclusionKind.Ellipsoid);
      }
    }
    /////////////////////////////////////////////////////////////////////////////////////////////
    public static Tensor GetPFactor(Tensor semiAxis, double E_matrix, double nu_matrix)  //for inclusion in isotropic matrix
    {

      Tensor P = new Tensor(6, 6);
      int sequence;
      double sigma;
      double eta;
      double G, K;

      G = E_matrix / (2 * (1 + nu_matrix));
      K = E_matrix / (3 * (1 - 2 * nu_matrix));

      sigma = (1 + nu_matrix) / (3 * (1 - nu_matrix));
      eta = 2 * (4 - 5 * nu_matrix) / (15 * (1 - nu_matrix));

      //Hill's moduli
      double k, l, n, m, p;

      k = -1 / (1 / G - 4 / E_matrix + 4 * nu_matrix * nu_matrix / E_matrix);
      l = 2 * k * nu_matrix;
      n = E_matrix + l * l / k;
      m = G;
      p = G;


      semiAxis = semiAxis.Sort(out sequence);				// a0>a1>a3

      if (semiAxis[0] == semiAxis[1] && semiAxis[0] == semiAxis[2])
      {
        P[0, 0] = P[1, 1] = P[2, 2] = (sigma / (3 * K) + eta / G) / 3;
        P[0, 1] = P[1, 0] = P[0, 2] = P[2, 0] = P[1, 2] = P[2, 1] = (sigma / (3 * K) - eta / (2 * G)) / 3;
        P[3, 3] = P[4, 4] = P[5, 5] = eta / G;
      }
      else if (semiAxis[0] == Constants.INFINITY)
      {
        if (semiAxis[1] == Constants.INFINITY)
        {
        }
        else if (semiAxis[1] == semiAxis[2])
        {
          P[1, 1] = P[2, 2] = (k + 4 * m) / (8 * m * (k + m));
          P[1, 2] = P[2, 1] = -k / (8 * m * (k + m));
          P[3, 3] = (k + 2 * m) / (2 * m * (k + m));
          P[4, 4] = P[5, 5] = 1 / (2 * p);
        }
      }
      else if (semiAxis[0] == semiAxis[1] && Functions.IsMuchMore(semiAxis[0], semiAxis[2]))
      {
        P[2, 2] = 1 / n;
        P[3, 3] = P[4, 4] = 1 / p;
      }
      else if (semiAxis[0] == semiAxis[1] && semiAxis[0] > semiAxis[2])
      {
      }
      else if (semiAxis[1] == semiAxis[2])
      {
      }
      else
      {
        Tensor S = new Tensor(3, 3, 3, 3);

        for (int i1 = 0; i1 < 3; i1++)
          for (int j1 = 0; j1 < 3; j1++)
            for (int k1 = 0; k1 < 3; k1++)
              for (int l1 = 0; l1 < 3; l1++)
                S[i1, j1, k1, l1] = (5 * nu_matrix - 1) / (15 * (1 - nu_matrix)) * Tensor.delta(i1, j1) * Tensor.delta(k1, l1) +
                                    (4 - 5 * nu_matrix) / (15 * (1 - nu_matrix)) * (Tensor.delta(i1, k1) * Tensor.delta(j1, l1) + Tensor.delta(i1, l1) * Tensor.delta(j1, k1));
      }

      // origin sequence
      P = P.Unsort(sequence);
      
      return (P);
    }
    public static Tensor GetEshelbyTensorIso(Tensor semiAxis, double nu_matrix)  //for inclusion in isotropic matrix
    {
      if (semiAxis.which_order != 1 || semiAxis.how_many_lines != 3)
      {
        throw new Exception("Elasticity: Inclusion: GetEshelbyTensorIso: wrong tensor!");
      }

      Tensor S = new Tensor(3, 3, 3, 3);
      Tensor S_matice = new Tensor(6, 6);
      int sequence;

      semiAxis = semiAxis.Sort(out sequence);				// a0>a1>a3

      if (semiAxis[0] == semiAxis[1])
      {
        if (semiAxis[0] == semiAxis[2])     //sphere
        {
          S[0, 0, 0, 0] = S[1, 1, 1, 1] = S[2, 2, 2, 2] = (7.0 - 5.0 * nu_matrix) / (15.0 * (1.0 - nu_matrix));
          S[0, 0, 1, 1] = S[1, 1, 2, 2] = S[2, 2, 0, 0] = S[0, 0, 2, 2] = S[1, 1, 0, 0] = S[2, 2, 1, 1] = (5.0 * nu_matrix - 1.0) / (15.0 * (1.0 - nu_matrix));
          S[0, 1, 0, 1] = S[1, 2, 1, 2] = S[2, 0, 2, 0] = (4.0 - 5.0 * nu_matrix) / (15.0 * (1.0 - nu_matrix));

          S = DoSymmetryOfS(S);
        }
        else if (semiAxis[0] > semiAxis[2])   
        {
          if (Functions.IsMuchMore(semiAxis[0], semiAxis[2]))    //penny shape
          {
            S[0, 0, 0, 0] = S[1, 1, 1, 1] = (13.0 - 8.0 * nu_matrix) / (32.0 * (1 - nu_matrix)) * Math.PI * semiAxis[2] / semiAxis[0];
            S[2, 2, 2, 2] = 1.0 - (1.0 - 2.0 * nu_matrix) / (1.0 - nu_matrix) * Math.PI / 4.0 * semiAxis[2] / semiAxis[0];
            S[0, 0, 1, 1] = S[1, 1, 0, 0] = (8.0 * nu_matrix - 1.0) / (32.0 * (1.0 - nu_matrix)) * Math.PI * semiAxis[2] / semiAxis[0];
            S[0, 0, 2, 2] = S[1, 1, 2, 2] = (2.0 * nu_matrix - 1.0) / (8.0 * (1.0 - nu_matrix)) * Math.PI * semiAxis[2] / semiAxis[0];
            S[2, 2, 0, 0] = S[2, 2, 1, 1] = nu_matrix / (1.0 - nu_matrix) * (1.0 - (4.0 * nu_matrix + 1) / (8.0 * nu_matrix) * Math.PI * semiAxis[2] / semiAxis[0]);
            S[0, 1, 0, 1] = (7.0 - 8.0 * nu_matrix) / (32.0 * (1.0 - nu_matrix)) * Math.PI * semiAxis[2] / semiAxis[0];
            S[0, 2, 0, 2] = S[1, 2, 1, 2] = 0.5 * (1.0 + (nu_matrix - 2.0) / (1.0 - nu_matrix) * Math.PI / 4.0 * semiAxis[2] / semiAxis[0]);

            S = DoSymmetryOfS(S);
          }
          else     //oblate spheroid
          {
            Tensor I = new Tensor(3);
            Tensor II = new Tensor(3, 3);

            I[0] = I[1] = 2.0 * Math.PI * semiAxis[0] * semiAxis[0] * semiAxis[2] / Math.Pow(semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2], 1.5) *
                          (Math.Acos(semiAxis[2] / semiAxis[0]) - semiAxis[2] / semiAxis[0] * Math.Sqrt(1 - semiAxis[2] * semiAxis[2] / (semiAxis[0] * semiAxis[0])));
            I[2] = 4 * Math.PI - 2.0 * I[0];

            II[0, 2] = (I[0] - I[2]) / (semiAxis[2] * semiAxis[2] - semiAxis[0] * semiAxis[0]);
            II[2, 2] = (4.0 * Math.PI / (semiAxis[2] * semiAxis[2]) - 2.0 * II[0, 2]) / 3.0;
            II[0, 0] = Math.PI / (semiAxis[0] * semiAxis[0]) - 0.25 * II[0, 2];

            II[2, 0] = II[1, 2] = II[2, 1] = II[0, 2];
            II[1, 1] = II[0, 1] = II[1, 0] = II[0, 0];

            S = EshelbyFromIntegrals(semiAxis, I, II, nu_matrix);
          }
        }
      }
      else if (semiAxis[0] == Constants.INFINITY)      //elliptic cylinder
      {
        S[0, 0, 0, 0] = 0.0;
        S[1, 1, 1, 1] = 1.0 / (2.0 * (1.0 - nu_matrix)) * ((semiAxis[2] * semiAxis[2] + 2.0 * semiAxis[1] * semiAxis[2]) / ((semiAxis[1] + semiAxis[2]) * (semiAxis[1] + semiAxis[2])) +
                        (1.0 - 2.0 * nu_matrix) * semiAxis[2] / (semiAxis[1] + semiAxis[2]));
        S[2, 2, 2, 2] = 1.0 / (2.0 * (1.0 - nu_matrix)) * ((semiAxis[1] * semiAxis[1] + 2.0 * semiAxis[1] * semiAxis[2]) / ((semiAxis[1] + semiAxis[2]) * (semiAxis[1] + semiAxis[2])) +
                        (1.0 - 2.0 * nu_matrix) * semiAxis[1] / (semiAxis[1] + semiAxis[2]));
        S[1, 1, 2, 2] = 1.0 / (2.0 * (1.0 - nu_matrix)) * ((semiAxis[2] * semiAxis[2]) / ((semiAxis[1] + semiAxis[2]) * (semiAxis[1] + semiAxis[2])) -
                        (1.0 - 2.0 * nu_matrix) * semiAxis[2] / (semiAxis[1] + semiAxis[2]));
        S[2, 2, 1, 1] = 1.0 / (2.0 * (1.0 - nu_matrix)) * ((semiAxis[1] * semiAxis[1]) / ((semiAxis[1] + semiAxis[2]) * (semiAxis[1] + semiAxis[2])) -
                        (1.0 - 2.0 * nu_matrix) * semiAxis[1] / (semiAxis[1] + semiAxis[2]));
        S[1, 1, 0, 0] = 1.0 / (2.0 * (1.0 - nu_matrix)) * 2.0 * nu_matrix * semiAxis[2] / (semiAxis[1] + semiAxis[2]);
        S[2, 2, 0, 0] = 1.0 / (2.0 * (1.0 - nu_matrix)) * 2.0 * nu_matrix * semiAxis[1] / (semiAxis[1] + semiAxis[2]);
        S[0, 0, 1, 1] = 0.0;
        S[0, 0, 2, 2] = 0.0;
        S[1, 2, 1, 2] = 1.0 / (2.0 * (1.0 - nu_matrix)) * ((semiAxis[1] * semiAxis[1] + semiAxis[2] * semiAxis[2]) / (2.0 * (semiAxis[1] + semiAxis[2]) * (semiAxis[1] + semiAxis[2])) +
                        (1.0 - 2.0 * nu_matrix) / 2.0);
        S[2, 0, 2, 0] = semiAxis[1] / (2.0 * (semiAxis[1] + semiAxis[2]));
        S[0, 1, 0, 1] = semiAxis[2] / (2.0 * (semiAxis[1] + semiAxis[2]));

        S = DoSymmetryOfS(S);
      }
      else if (semiAxis[0] > semiAxis[1])
      {
        if (semiAxis[1] == semiAxis[2])   //prolate spheroid
        {
          Tensor I = new Tensor(3);
          Tensor II = new Tensor(3, 3);

          I[1] = I[2] = 2.0 * Math.PI * semiAxis[0] * semiAxis[2] * semiAxis[2] / Math.Pow(semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2], 1.5) *
                        (semiAxis[0] / semiAxis[2] * Math.Sqrt(semiAxis[0] * semiAxis[0] / (semiAxis[2] * semiAxis[2]) - 1.0) - Functions.ACosh(semiAxis[0] / semiAxis[2]));
          I[0] = 4.0 * Math.PI - 2.0 * I[1];

          II[0, 1] = (I[1] - I[0]) / (semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]);
          II[0, 0] = (4 * Math.PI / (semiAxis[0] * semiAxis[0]) - 2.0 * II[0, 1]) / 3.0;
          II[1, 1] = Math.PI / (semiAxis[1] * semiAxis[1]) - (I[1] - I[0]) / (4.0 * (semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]));

          II[1, 0] = II[0, 2] = II[2, 0] = II[0, 1];
          II[2, 2] = II[1, 2] = II[2, 1] = II[1, 1];

          S = EshelbyFromIntegrals(semiAxis, I, II, nu_matrix);
        }
        else if (semiAxis[1] > semiAxis[2])
        {
          double arg;
          double F_int, E_int;
          Tensor I = new Tensor(3);
          Tensor II = new Tensor(3, 3);

          if (Functions.IsMuchMore(semiAxis[1], semiAxis[2]))     //flat ellipsoid
          {
            arg = ((semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]) / (semiAxis[0] * semiAxis[0]));

            F_int = Elliptic.ellipticintegralk(arg);
            E_int = Elliptic.ellipticintegrale(arg);

            I[0] = 4.0 * Math.PI * semiAxis[1] * semiAxis[2] * (F_int - E_int) / (semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]);
            I[1] = 4.0 * Math.PI * (semiAxis[2] * E_int / semiAxis[1] - semiAxis[1] * semiAxis[2] * (F_int - E_int) / (semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]));
            I[2] = 4.0 * Math.PI * (1.0 - semiAxis[2] * E_int / semiAxis[1]);
          }
          else   //general ellipsoid
          {
            double theta;

            theta = Math.Asin(Math.Sqrt(1 - semiAxis[2] * semiAxis[2] / (semiAxis[0] * semiAxis[0])));
            arg = ((semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]) / (semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2]));

            F_int = Elliptic.incompleteellipticintegralk(theta, arg);
            E_int = Elliptic.incompleteellipticintegrale(theta, arg);

            I[0] = 4.0 * Math.PI * semiAxis[0] * semiAxis[1] * semiAxis[2] / ((semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]) * Math.Sqrt(semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2])) *
                   (F_int - E_int);
            I[2] = 4.0 * Math.PI * semiAxis[0] * semiAxis[1] * semiAxis[2] / ((semiAxis[1] * semiAxis[1] - semiAxis[2] * semiAxis[2]) * Math.Sqrt(semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2])) *
                   (semiAxis[1] * Math.Sqrt(semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2]) / (semiAxis[0] * semiAxis[2]) - E_int);
            I[1] = 4.0 * Math.PI - I[0] - I[2];
          }

          II[0, 1] = (I[1] - I[0]) / (semiAxis[0] * semiAxis[0] - semiAxis[1] * semiAxis[1]);
          II[0, 2] = (I[2] - I[0]) / (semiAxis[0] * semiAxis[0] - semiAxis[2] * semiAxis[2]);
          II[0, 0] = (4.0 * Math.PI / (semiAxis[0] * semiAxis[0]) - II[0, 1] - II[0, 2]) / 3.0;

          // 1st cyclic permutation
          II[1, 2] = (I[2] - I[1]) / (semiAxis[1] * semiAxis[1] - semiAxis[2] * semiAxis[2]);
          II[1, 0] = II[0, 1];
          II[1, 1] = (4.0 * Math.PI / (semiAxis[1] * semiAxis[1]) - II[1, 2] - II[1, 0]) / 3.0;

          // 2nd cyclic permutation
          II[2, 0] = II[0, 2];
          II[2, 1] = II[1, 2];
          II[2, 2] = (4.0 * Math.PI / (semiAxis[2] * semiAxis[2]) - II[2, 0] - II[2, 1]) / 3.0;

          S = EshelbyFromIntegrals(semiAxis, I, II, nu_matrix);

        }
      }
      else
        throw new Exception("Elasticity: Inclusion: GetEshelbyTensor:wrong a!!!");

      S_matice = S.Conversion4thTo2ndOrderEshelby();
      S_matice = S_matice.Unsort(sequence);

      return (S_matice);
    }
    public static Tensor GetEshelbyTensorAniso(Tensor semiAxis, Tensor L0, int M_partition, int N_partition)
    {
      if (semiAxis.which_order != 1 || semiAxis.how_many_lines != 3 ||
          L0.which_order != 4 || L0.how_many_lines != 3 || L0.how_many_columns != 3 ||
          L0.how_many_lines2 != 3 || L0.how_many_columns2 != 3)
      {
        throw new Exception("Elasticity: Inclusion: GetEshelbyTensorAniso: wrong tensor!");
      }

      Tensor xiBar = new Tensor(3);
      Tensor zeta = new Tensor(3);
      Tensor K = new Tensor(3, 3);
      Tensor N = new Tensor(3, 3);
      Tensor epsilon = new Tensor(3, 3, 3);
      double[, , , , ,] G = new double[3, 3, 3, 3, M_partition, N_partition];
      Tensor S = new Tensor(3, 3, 3, 3);
      double D;

      Tensor w1 = new Tensor(N_partition);
      Tensor x1 = new Tensor(N_partition);
      Quadrature.GauLegF(0, 2 * Math.PI, x1, w1);
      Tensor w2 = new Tensor(M_partition);
      Tensor x2 = new Tensor(M_partition);
      Quadrature.GauLegF(-1, 1, x2, w2);


      //initialization
      epsilon = Tensor.PermutationTensor();

      for (int p = 0; p < M_partition; p++)
        for (int q = 0; q < N_partition; q++)
        {
          zeta[0] = Math.Sqrt(1.0 - x2[p] * x2[p]) * Math.Cos(x1[q]);
          zeta[1] = Math.Sqrt(1.0 - x2[p] * x2[p]) * Math.Sin(x1[q]);
          zeta[2] = x2[p];

          xiBar[0] = zeta[0] / semiAxis[0];
          xiBar[1] = zeta[1] / semiAxis[1];
          xiBar[2] = zeta[2] / semiAxis[2];

          K.Clear();
          for (int i = 0; i < 3; i++)
            for (int k = 0; k < 3; k++)
              for (int j = 0; j < 3; j++)
                for (int l = 0; l < 3; l++)
                  K[i, k] += L0[i, j, k, l] * xiBar[j] * xiBar[l];

          D = 0;
          N.Clear();
          for (int i = 0; i < 3; i++)
            for (int j = 0; j < 3; j++)
              for (int k = 0; k < 3; k++)
              {
                D += epsilon[i, j, k] * K[i, 0] * K[j, 1] * K[k, 2];
                for (int l = 0; l < 3; l++)
                  for (int m = 0; m < 3; m++)
                    for (int n = 0; n < 3; n++)
                      N[i, j] += 0.5 * epsilon[i, k, l] * epsilon[j, m, n] * K[k, m] * K[l, n];
              }

           for (int i = 0; i < 3; i++)
             for (int j = 0; j < 3; j++)
               for (int k = 0; k < 3; k++)
                 for (int l = 0; l < 3; l++)
                   G[i, j, k, l, p, q] += xiBar[k] * xiBar[l] * N[i, j] / D;
                
        }

      for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
          for (int k = 0; k < 3; k++)
            for (int l = 0; l < 3; l++)
              for (int m = 0; m < 3; m++)
                for (int n = 0; n < 3; n++)
                  for (int p = 0; p < M_partition; p++)
                    for (int q = 0; q < N_partition; q++)
                      S[i, j, k, l] += 1.0 / (8.0 * Math.PI) * L0[m, n, k, l] * (G[i, m, j, n, p, q] + G[j, m, i, n, p, q]) * w1[q] * w2[p];


      return (S.Conversion4thTo2ndOrderEshelby());
    }
    private static Tensor EshelbyFromIntegrals(Tensor semiAxis, Tensor I, Tensor II, double nu_matrix)
    {
      if (semiAxis.which_order != 1 || semiAxis.how_many_lines != 3 || I.which_order != 1 || semiAxis.how_many_lines != 3 ||
          II.which_order != 2 || II.how_many_lines != 3 || II.how_many_columns != 3)
      {
        throw new Exception("Elasticity: Inclusion: EshelbyFromIntegrals: wrong tensor!");
      }

      Tensor S = new Tensor(3, 3, 3, 3);
      double Q, R;
      int i = 0, k = 0, l = 0;

      Q = 3.0 / (8.0 * Math.PI * (1.0 - nu_matrix));
      R = (1.0 - 2.0 * nu_matrix) / (8.0 * Math.PI * (1.0 - nu_matrix));

      for (i = 0; i < 3; i++)
      {
        if (i == 0)
        {
          k = 1;
          l = 2;
        }
        else if (i == 1)
        {
          k = 2;
          l = 0;
        }
        else if (i == 2)
        {
          k = 0;
          l = 1;
        }

        S[i, i, i, i] = Q * semiAxis[i] * semiAxis[i] * II[i, i] + R * I[i];
        S[i, i, k, k] = Q / 3.0 * semiAxis[k] * semiAxis[k] * II[i, k] - R * I[i];
        S[i, i, l, l] = Q / 3.0 * semiAxis[l] * semiAxis[l] * II[i, l] - R * I[i];
        S[i, k, i, k] = Q / 6.0 * (semiAxis[i] * semiAxis[i] + semiAxis[k] * semiAxis[k]) * II[i, k] + R / 2.0 * (I[i] + I[k]);
      }

      S = DoSymmetryOfS(S);

      return S;
    }

    private static Tensor DoSymmetryOfS(Tensor S)
    {
      if (S.which_order != 4 || S.how_many_lines != 3 || S.how_many_lines2 != 3 ||
          S.how_many_columns != 3 || S.how_many_columns2 != 3)
      {
        throw new Exception("Elasticity: Inclusion: DoSymmetryOfS: wrong tensor!");
      }

      Tensor a = new Tensor(S);

      for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
          for (int k = 0; k < 3; k++)
            for (int l = 0; l < 3; l++)
            {
              if (a[i, j, k, l] != 0)
                a[i, j, l, k] = a[j, i, k, l] = a[j, i, l, k] = a[i, j, k, l];
              else if (a[j, i, k, l] != 0)
                a[i, j, l, k] = a[i, j, k, l] = a[j, i, l, k] = a[j, i, k, l];
              else if (a[i, j, l, k] != 0)
                a[i, j, k, l] = a[j, i, k, l] = a[j, i, l, k] = a[i, j, l, k];
              else if (a[j, i, l, k] != 0)
                a[i, j, k, l] = a[j, i, k, l] = a[i, j, l, k] = a[j, i, l, k];
            }

      return a;
    }

    public override string ToString()
    {
      return name;
    }
    public void ClearInclusion()
    {
      base.ClearConstituent();
      _a[0] = _a[1] = _a[2] = 1;
      this._EulerAngles.Clear();
      this._VolumeFrac = 0;
      this.name = "";
      this._RandomOrientation = false;
    }

    public XmlElement ToXml(XmlDocument xmldoc)
    {
      XmlElement _incl;
      XmlElement _text1;
      XmlAttribute _atrib;

      _incl = xmldoc.CreateElement("inclusion");

      _atrib = xmldoc.CreateAttribute("name");
      _atrib.Value = name;
      _incl.SetAttributeNode(_atrib);

      _text1 = xmldoc.CreateElement("vol_frac");
      _text1.InnerText = Convert.ToString(VolFrac, Constants.Separators.FormatProvider);
      _incl.AppendChild(_text1);

      _incl.AppendChild(SemiAxes.ToXml(xmldoc, "semi_axes"));

      _incl.AppendChild(EulerAngles.ToXml(xmldoc, "Euler_angles"));

      _incl.AppendChild(StiffnessMatrix.ToXml(xmldoc, "stiffness_matrix"));

      return (_incl);
    }
    public void FromXml(XmlNode xmlnode)
    {
      string name = "";
      this.ClearInclusion();
      int[] count = new int[4];

      count[0] = count[1] = count[2] = count[3] = 0;

      this.name = "new inclusion";
      for (int i = 0; i < xmlnode.Attributes.Count; i++)
      {
        if (xmlnode.Attributes[i].Name == "name")
        {
          this.name = xmlnode.Attributes[i].Value;
        }
      }

      if (!(xmlnode.ChildNodes.Count == 4))
      {
        throw new Exception("xml: inclusion: wrong number of children!!!");
      }
      else
      {
        for (int i = 0; i < 4; i++)
        {
          switch (xmlnode.ChildNodes[i].Name)
          {
            case "vol_frac":
              count[0]++;

              this._VolumeFrac = Functions.StringToDouble(xmlnode.ChildNodes[i].InnerText, "xml: inclusion: wrong vol_frac!!!");
              break;

            case "tensor":
              for (int j = 0; j < xmlnode.ChildNodes[i].Attributes.Count; j++)
              {
                if (xmlnode.ChildNodes[i].Attributes[j].Name == "name")
                {
                  name = xmlnode.ChildNodes[i].Attributes[j].Value;
                }
              }
              switch (name)
              {
                case "semi_axes":
                  count[1]++;
                  this.SemiAxes.FromXml(xmlnode.ChildNodes[i]);
                  break;

                case "Euler_angles":
                  count[2]++;
                  this.EulerAngles.FromXml(xmlnode.ChildNodes[i]);
                  break;

                case "stiffness_matrix":
                  count[3]++;
                  this.StiffnessMatrix.FromXml(xmlnode.ChildNodes[i]);
                  break;

                default:
                  throw new Exception("xml: inclusion: wrong tensor name!!!");
              }
              break;

            default:
              throw new Exception("xml: inclusion: wrong tensor child!!!");
          }

        }
        if (!(count[0] == 1 && count[1] == 1 && count[2] == 1 && count[3] == 1))
        {
          throw new Exception("xml: inclusion: there miss some inclusion parameters!!!");
        }

      }
    }

  }
}
