﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;

namespace CVUT.Vorel
{
  public class DataXY
  {
    public List<double> x;
    public List<double> y;
    public string xLabel = "";
    public string yLabel = "";
    public string name;
    public Color color;

    public DataXY()
    {
      x = new List<double>();
      y = new List<double>();
      this.color = Color.Gray; // Default color
    }
    public DataXY(List<double> x, List<double> y) : this()
    {
      this.x = x;
      this.y = y;
      CheckDataLength();
    }
    public DataXY(double[] x, double[] y) : this()
    {
      foreach(double value in x)
        this.x.Add(value);

      foreach (double value in y)
        this.y.Add(value);

      CheckDataLength();
    }
    public DataXY(DataXY xy)
    {
      x = xy.x.ToList();
      y = xy.y.ToList();
      name = xy.name;
      xLabel = xy.xLabel;
      yLabel = xy.yLabel;
    }
    public DataXY(FileInfo fileInfo) : this()
    {
      ReadFromFile(fileInfo);
      CheckDataLength();
    }
    public DataXY(string fileName)
      : this()
    {
      ReadFromFile(fileName);
      CheckDataLength();
    }

    private void CheckDataLength()
    {
      if (x.Count != y.Count)
        throw new Exception("Different data counts in DataXY");
    }
    public void Clear()
    {
      x.Clear();
      y.Clear();
    }
    public int Count
    {
      get { return x.Count; }
    }
    public bool OnlyNumbers()
    {
      for (int i = 0; i < this.Count; i++)
      {
        if (Double.IsNaN(x[i]) || Double.IsNaN(y[i]))
        {
          return false;
        }
      }
      return true;
    }
    public bool OnlyRealNumbers()
    {
      for (int i = 0; i < this.Count; i++)
      {
        if (Double.IsNaN(x[i]) || Double.IsNaN(y[i]) ||
            Double.IsPositiveInfinity(Math.Abs(x[i])) || Double.IsPositiveInfinity(Math.Abs(y[i])))
        {
          return false;
        }
      }
      return true;
    }
    public DataXY Collection(int fromIndex, int toIndex)
    {
      DataXY dummy = new DataXY();

      dummy.name = this.name;
      dummy.xLabel = this.xLabel;
      dummy.yLabel = this.yLabel;

      if (this.Count == 0)
      {
        return (dummy);
      }
      if (fromIndex < 0)
      {
        fromIndex = 0;
      }
      if (toIndex > this.Count - 1)
      {
        toIndex = this.Count - 1;
      }

      if (fromIndex > toIndex)
      {
        throw new Exception("DataXY: Collection: wrong indexes");
      }

      for (int i = fromIndex; i <= toIndex; i++)
      {
        dummy.Add(this.x[i], this.y[i]);
      }

      return (dummy);
    }
    private static bool ControlListDataXY(List<DataXY> dataList)
    {
      for (int i = 0; i < dataList.Count; i++)
      {
        if (dataList[i].x.Count != dataList[i].y.Count)
          return false;
        if (dataList[0].Count != dataList[i].Count)
          return false;
      }
      return true;
    }

    public double Xmax
    {
      get { return Common.Max(x); }
    }
    public double Xmin
    {
      get { return Common.Min(x); }
    }
    public double Ymax
    {
      get { return Common.Max(y); }
    }
    public double Ymin
    {
      get { return Common.Min(y); }
    }
    public double DeltaX
    {
      get
      {
        if (IsEquidistant)
          return x[1] - x[0];
        else
          return Double.NaN;
      }
    }
    public void Add(double x, double y)
    {
      this.x.Add(x);
      this.y.Add(y);
    }

    private void ReadFromFile(FileInfo fileInfo)
    {
      FileStream fs = new FileStream(fileInfo.FullName, FileMode.Open);
      StreamReader sr = new StreamReader(fs);
      string[] items_in_line = new string[2]; // Two numbers in one line
      string line;

      name = fileInfo.FullName;
      color = Color.DarkBlue;

      // Reading header line if any
      //line = sr.ReadLine();

      x.Clear();
      y.Clear();
      
      while ((line = sr.ReadLine()) != null)
      {
        if (line != null)
        {
          // Splitting line
          items_in_line = line.Split(new Char[] { ',' });
          x.Add(Double.Parse(items_in_line[0]));
          y.Add(Double.Parse(items_in_line[1]));
        }
      }
      fs.Close();
      sr.Close();
    }
    public void ReadFromFile(string fileName)
    {
      FileStream fs = new FileStream(fileName, FileMode.Open);
      StreamReader sr = new StreamReader(fs);
      string[] items_in_line = new string[2]; // Two numbers in one line
      string line;
      double dummy;

      // Reading header line if any
      //line = sr.ReadLine();
      this.Clear();
      while ((line = sr.ReadLine()) != null)
      {
        if (line != null)
        {
          // Splitting line
          items_in_line = Common.SplittingLine(line, 2);

          // Adding the line
          if (Double.TryParse(items_in_line[0].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
            x.Add(dummy);
          if (Double.TryParse(items_in_line[1].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
            y.Add(dummy);

        }
      }
      fs.Close();
      sr.Close();
    }
    public static List<DataXY> ReadFromFile(string fileName, int nItems)
    {
      if(nItems<1)
      {
        throw new Exception("DataXY: ReadFromFile(string fileName, int nItems): nItems<1!!!");
      }

      FileStream fs = new FileStream(fileName, FileMode.Open);
      StreamReader sr = new StreamReader(fs);
      string line;
      double dummy;
      List<DataXY> dummyList;

      // Reading header line if any
      //line = sr.ReadLine();

      string[] items_in_line = new string[nItems]; 
      dummyList = new List<DataXY>();
      for (int i = 0; i < nItems - 1; i++)
      {
        dummyList.Add(new DataXY());
      }
      while ((line = sr.ReadLine()) != null)
      {
        if (line != null)
        {
          // Splitting line
          items_in_line = Common.SplittingLine(line, nItems);

          switch (nItems)
          {
            case 2:
              {
                // Adding the line
                if (Double.TryParse(items_in_line[0].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
                  dummyList[0].x.Add(dummy);
                if (Double.TryParse(items_in_line[1].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
                  dummyList[0].y.Add(dummy);
                break;
              }
            case 3:     //pro triaxial se zadanim deformaci i v pricnych smerech
              {
                if (Double.TryParse(items_in_line[0].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
                  dummyList[0].x.Add(dummy);    //deformace

                if (Double.TryParse(items_in_line[2].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
                  dummyList[0].y.Add(dummy);    //napeti


                if (Double.TryParse(items_in_line[1].Replace(CVUT.Vorel.Constants.Separators.BadDecimalSeparator, CVUT.Vorel.Constants.Separators.DecimalSeparator), System.Globalization.NumberStyles.Float, CVUT.Vorel.Constants.Separators.FormatProvider, out dummy))
                {
                  dummyList[1].x.Add(dummy);    //deformace, v druhem smeru
                  dummyList[1].y.Add(dummyList[0].y[0]);
                }
                break;
              }
            default:
              {
                throw new Exception("DataXY: ReadFromFile(string fileName, int nItems): wrong case!!!");
              }
          }
        }
      }
      fs.Close();
      sr.Close();

      if (!(ControlListDataXY(dummyList)))
        throw new Exception("DataXY: ReadFromFile(string fileName, int nItems): wrong data count!!!");

      return (dummyList);
    }

    public void WriteToFile(string fileName)
    {
      string text = "";

      using (StreamWriter sw = File.CreateText(fileName))
      {
        for (int i = 0; i < this.Count; i++)
        {
          text += Convert.ToString(this.x[i], Constants.Separators.FormatProvider) + "\t" 
                  + Convert.ToString(this.y[i], Constants.Separators.FormatProvider) + "\n";
        }

        sw.Write(text);
      }
    }
    public static void WriteToFile(string fileName, List<DataXY> dataList, int nItems)
    {
      if (!(ControlListDataXY(dataList)))
        throw new Exception("DataXY: WriteToFile(string fileName, List<DataXY> dataList, Constants.ETestKind type): wrong data count!!!");

      string text = "";

      using (StreamWriter sw = File.CreateText(fileName))
      {
        switch (nItems)
        {
          case 2:
            {
              for (int i = 0; i < dataList[0].Count; i++)
              {
                text += Convert.ToString(dataList[0].x[i], Constants.Separators.FormatProvider) + "\t"
                        + Convert.ToString(dataList[0].y[i], Constants.Separators.FormatProvider) + "\n";
              }
              break;
            }
          case 3: //pro triaxial s ukladanim i pricnych deformaci
            {
              for (int i = 0; i < dataList[0].Count; i++)
              {
                text += Convert.ToString(dataList[0].x[i], Constants.Separators.FormatProvider) + "\t"
                        + Convert.ToString(dataList[1].x[i], Constants.Separators.FormatProvider) + "\t"
                        + Convert.ToString(dataList[0].y[i], Constants.Separators.FormatProvider) + "\n";
              }
              break;
            }
          default:
            {
              throw new Exception("DataXY: WriteToFile: wrong case!!!");
            }
        }

        sw.Write(text);
      }
    }

    public bool IsEquidistant
    {
      get
      {
        double deltaX = x[1] - x[0];

        // Check if the time series is equidistant
        for (int i = 1; i < x.Count; i++)
        {
          if (x[i] - x[i - 1] != deltaX)
          {
            return false;
          }
        }
        return true;
      }
    }
    public void Log(ILog log)
    {
      for (int i = 0; i < Count; i++)
        log.Write(x[i].ToString() + "\t" + y[i].ToString());
    }
    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      sb.AppendLine(name);
      sb.AppendLine();
      for (int i = 0; i < Count; i++)
        sb.AppendLine(x[i].ToString() + "\t" + y[i].ToString());
      return sb.ToString();
    }
    //public override string ToString()
    //{
    //  return name;
    //}
  }
}
