﻿using System;
using System.Collections.Generic;
using System.Text;

namespace CVUT.Vorel
{
  class Elliptic
  {
    private const double machineepsilon = 5E-16;

    private static int ifloor(double x)
    {
      return (int)Math.Floor(x);
    }

    /*************************************************************************
Cephes Math Library Release 2.8:  June, 2000
Copyright 1984, 1987, 1995, 2000 by Stephen L. Moshier

Contributors:
    * Sergey Bochkanov (ALGLIB project). Translation from C to
      pseudocode.

See subroutines comments for additional copyrights.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.

- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************/


    /*************************************************************************
    Complete elliptic integral of the first kind

    Approximates the integral



               pi/2
                -
               | |
               |           dt
    K(m)  =    |    ------------------
               |                        2
             | |    Math.Sqrt( 1 - m sin t )
              -
               0

    using the approximation

        P(x)  -  log x Q(x).

    ACCURACY:

                         Relative error:
    arithmetic   domain     # trials      peak         rms
       IEEE       0,1        30000       2.5e-16     6.8e-17

    Cephes Math Library, Release 2.8:  June, 2000
    Copyright 1984, 1987, 2000 by Stephen L. Moshier
    *************************************************************************/
    public static double ellipticintegralk(double m)
    {
      double result;

      result = ellipticintegralkhighprecision(1.0 - m);
      return result;
    }


    /*************************************************************************
    Complete elliptic integral of the first kind

    Approximates the integral



               pi/2
                -
               | |
               |           dt
    K(m)  =    |    ------------------
               |                        2
             | |    Math.Sqrt( 1 - m sin t )
              -
               0

    where m = 1 - m1, using the approximation

        P(x)  -  log x Q(x).

    The argument m1 is used rather than m so that the logarithmic
    singularity at m = 1 will be shifted to the origin; this
    preserves maximum accuracy.

    K(0) = pi/2.

    ACCURACY:

                         Relative error:
    arithmetic   domain     # trials      peak         rms
       IEEE       0,1        30000       2.5e-16     6.8e-17

    Ŕëăîđčňě âç˙ň čç áčáëčîňĺęč Cephes
    *************************************************************************/
    public static double ellipticintegralkhighprecision(double m1)
    {
      double result;
      double p;
      double q;

      if (m1 <= machineepsilon)
      {
        result = 1.3862943611198906188E0 - 0.5 * Math.Log(m1);
      }
      else
      {
        p = 1.37982864606273237150E-4;
        p = p * m1 + 2.28025724005875567385E-3;
        p = p * m1 + 7.97404013220415179367E-3;
        p = p * m1 + 9.85821379021226008714E-3;
        p = p * m1 + 6.87489687449949877925E-3;
        p = p * m1 + 6.18901033637687613229E-3;
        p = p * m1 + 8.79078273952743772254E-3;
        p = p * m1 + 1.49380448916805252718E-2;
        p = p * m1 + 3.08851465246711995998E-2;
        p = p * m1 + 9.65735902811690126535E-2;
        p = p * m1 + 1.38629436111989062502E0;
        q = 2.94078955048598507511E-5;
        q = q * m1 + 9.14184723865917226571E-4;
        q = q * m1 + 5.94058303753167793257E-3;
        q = q * m1 + 1.54850516649762399335E-2;
        q = q * m1 + 2.39089602715924892727E-2;
        q = q * m1 + 3.01204715227604046988E-2;
        q = q * m1 + 3.73774314173823228969E-2;
        q = q * m1 + 4.88280347570998239232E-2;
        q = q * m1 + 7.03124996963957469739E-2;
        q = q * m1 + 1.24999999999870820058E-1;
        q = q * m1 + 4.99999999999999999821E-1;

        result = p - q * Math.Log(m1);
      }
      return result;
    }


    /*************************************************************************
    Incomplete elliptic integral of the first kind F(phi|m)

    Approximates the integral



                   phi
                    -
                   | |
                   |           dt
    F(phi_\m)  =    |    ------------------
                   |                        2
                 | |    Math.Sqrt( 1 - m sin t )
                  -
                   0

    of amplitude phi and modulus m, using the arithmetic -
    geometric mean algorithm.




    ACCURACY:

    Tested at random points with m in [0, 1] and phi as indicated.

                         Relative error:
    arithmetic   domain     # trials      peak         rms
       IEEE     -10,10       200000      7.4e-16     1.0e-16

    Cephes Math Library Release 2.8:  June, 2000
    Copyright 1984, 1987, 2000 by Stephen L. Moshier
    *************************************************************************/
    public static double incompleteellipticintegralk(double phi, double m)
{
    double result;
    double a;
    double b;
    double c;
    double e;
    double temp;
    double pio2;
    double t;
    double k;
    int d;
    int md;
    int s;
    int npio2;

    pio2 = 1.57079632679489661923;
    if( m==0 )
    {
        result = phi;
        return result;
    }
    a = 1-m;
    if( a==0 )
    {
        result = Math.Log(Math.Tan(0.5*(pio2+phi)));
        return result;
    }
    npio2 = ifloor(phi/pio2);
    if( npio2%2!=0 )
    {
        npio2 = npio2+1;
    }
    if( npio2!=0 )
    {
        k = ellipticintegralk(1-a);
        phi = phi-npio2*pio2;
    }
    else
    {
        k = 0;
    }
    if( phi<0 )
    {
        phi = -phi;
        s = -1;
    }
    else
    {
        s = 0;
    }
    b = Math.Sqrt(a);
    t = Math.Tan(phi);
    if( Math.Abs(t)>10 )
    {
        e = 1.0/(b*t);
        if( Math.Abs(e)<10 )
        {
            e = Math.Atan(e);
            if( npio2==0 )
            {
                k = ellipticintegralk(1-a);
            }
            temp = k-incompleteellipticintegralk(e, m);
            if( s<0 )
            {
                temp = -temp;
            }
            result = temp+npio2*k;
            return result;
        }
    }
    a = 1.0;
    c = Math.Sqrt(m);
    d = 1;
    md = 0;
    while(Math.Abs(c/a)>machineepsilon)
    {
        temp = b/a;
        phi = phi+Math.Atan(t*temp)+md*Math.PI;
        md = (int)Math.Truncate((phi+pio2)/Math.PI);
        t = t*(1.0+temp)/(1.0-temp*t*t);
        c = 0.5*(a-b);
        temp = Math.Sqrt(a*b);
        a = 0.5*(a+b);
        b = temp;
        d = d+d;
    }
    temp = (Math.Atan(t)+md*Math.PI)/(d*a);
    if( s<0 )
    {
        temp = -temp;
    }
    result = temp+npio2*k;
    return result;
}


    /*************************************************************************
    Complete elliptic integral of the second kind

    Approximates the integral


               pi/2
                -
               | |                      2
    E(m)  =    |    Math.Sqrt( 1 - m sin t ) dt
             | |
              -
               0

    using the approximation

         P(x)  -  x log x Q(x).

    ACCURACY:

                         Relative error:
    arithmetic   domain     # trials      peak         rms
       IEEE       0, 1       10000       2.1e-16     7.3e-17

    Cephes Math Library, Release 2.8: June, 2000
    Copyright 1984, 1987, 1989, 2000 by Stephen L. Moshier
    *************************************************************************/
    public static double ellipticintegrale(double m)
    {
      double result;
      double p;
      double q;

      if (!(m >= 0 && m <= 1))
        throw new Exception("Elliptic: ellipticintegrale");
      m = 1 - m;
      if (m == 0)
      {
        result = 1;
        return result;
      }
      p = 1.53552577301013293365E-4;
      p = p * m + 2.50888492163602060990E-3;
      p = p * m + 8.68786816565889628429E-3;
      p = p * m + 1.07350949056076193403E-2;
      p = p * m + 7.77395492516787092951E-3;
      p = p * m + 7.58395289413514708519E-3;
      p = p * m + 1.15688436810574127319E-2;
      p = p * m + 2.18317996015557253103E-2;
      p = p * m + 5.68051945617860553470E-2;
      p = p * m + 4.43147180560990850618E-1;
      p = p * m + 1.00000000000000000299E0;
      q = 3.27954898576485872656E-5;
      q = q * m + 1.00962792679356715133E-3;
      q = q * m + 6.50609489976927491433E-3;
      q = q * m + 1.68862163993311317300E-2;
      q = q * m + 2.61769742454493659583E-2;
      q = q * m + 3.34833904888224918614E-2;
      q = q * m + 4.27180926518931511717E-2;
      q = q * m + 5.85936634471101055642E-2;
      q = q * m + 9.37499997197644278445E-2;
      q = q * m + 2.49999999999888314361E-1;
      result = p - q * m * Math.Log(m);
      return result;
    }


    /*************************************************************************
    Incomplete elliptic integral of the second kind

    Approximates the integral


                   phi
                    -
                   | |
                   |                         2
    E(phi_\m)  =    |    Math.Sqrt( 1 - m sin t ) dt
                   |
                 | |
                  -
                   0

    of amplitude phi and modulus m, using the arithmetic -
    geometric mean algorithm.

    ACCURACY:

    Tested at random arguments with phi in [-10, 10] and m in
    [0, 1].
                         Relative error:
    arithmetic   domain     # trials      peak         rms
       IEEE     -10,10      150000       3.3e-15     1.4e-16

    Cephes Math Library Release 2.8:  June, 2000
    Copyright 1984, 1987, 1993, 2000 by Stephen L. Moshier
    *************************************************************************/
    public static double incompleteellipticintegrale(double phi, double m)
    {
      double result;
      double pio2;
      double a;
      double b;
      double c;
      double e;
      double temp;
      double lphi;
      double t;
      double ebig;
      int d;
      int md;
      int npio2;
      int s;

      pio2 = 1.57079632679489661923;
      if (m == 0)
      {
        result = phi;
        return result;
      }
      lphi = phi;
      npio2 = ifloor(lphi / pio2);
      if (npio2 % 2 != 0)
      {
        npio2 = npio2 + 1;
      }
      lphi = lphi - npio2 * pio2;
      if (lphi < 0)
      {
        lphi = -lphi;
        s = -1;
      }
      else
      {
        s = 1;
      }
      a = 1.0 - m;
      ebig = ellipticintegrale(m);
      if (a == 0)
      {
        temp = Math.Sin(lphi);
        if (s < 0)
        {
          temp = -temp;
        }
        result = temp + npio2 * ebig;
        return result;
      }
      t = Math.Tan(lphi);
      b = Math.Sqrt(a);

      //
      // Thanks to Brian Fitzgerald <fitzgb@mml0.meche.rpi.edu>
      // for pointing out an instability near odd multiples of pi/2
      //
      if (Math.Abs(t) > 10)
      {

        //
        // Transform the amplitude
        //
        e = 1.0 / (b * t);

        //
        // ... but avoid multiple recursions.
        //
        if (Math.Abs(e) < 10)
        {
          e = Math.Atan(e);
          temp = ebig + m * Math.Sin(lphi) * Math.Sin(e) - incompleteellipticintegrale(e, m);
          if (s < 0)
          {
            temp = -temp;
          }
          result = temp + npio2 * ebig;
          return result;
        }
      }
      c = Math.Sqrt(m);
      a = 1.0;
      d = 1;
      e = 0.0;
      md = 0;
      while (Math.Abs(c / a) > machineepsilon)
      {
        temp = b / a;
        lphi = lphi + Math.Atan(t * temp) + md * Math.PI;
        md = (int)Math.Truncate((lphi + pio2) / Math.PI);
        t = t * (1.0 + temp) / (1.0 - temp * t * t);
        c = 0.5 * (a - b);
        temp = Math.Sqrt(a * b);
        a = 0.5 * (a + b);
        b = temp;
        d = d + d;
        e = e + c * Math.Sin(lphi);
      }
      temp = ebig / ellipticintegralk(m);
      temp = temp * ((Math.Atan(t) + md * Math.PI) / (d * a));
      temp = temp + e;
      if (s < 0)
      {
        temp = -temp;
      }
      result = temp + npio2 * ebig;
      return result;

    }

  }
}
