/*      P147 4.5 in Numerical Recipes in C                      */

/****************************************************
*                                                   *               
* FILE: 	==== gaussQ.cpp ====					            *   
*                                                   *
* Contains function for gauss quadrature            *
* with them                                         *
*                                                   *
* Authors:           Jan Vorel			                *
* Latest update:     2008							              *
*                                                   *
*****************************************************/


using System;
using System.Collections.Generic;
using System.Text;

namespace CVUT.Vorel
{
  class Quadrature
  {
    public static void GauLegF(double a, double b, Tensor x, Tensor w)
    {
      /* compute x(i) and w(i)  i=1,n  Legendre ordinates and weights */
      /* on interval  -1.0 to 1.0 (length is 2.0)                     */
      /* use ordinates and weights for Gauss Legendre integration     */

      if (x.which_order != 1 || w.which_order != 1 || x.how_many_lines != w.how_many_lines)
      {
        throw new Exception("GaussQ: GauLegF: wrong tensor!!!");
      }

      int i, j, m;
      double p1, p2, p3, pp, xl, xm, z, z1;
      int n = x.how_many_lines;

      x.Clear();
      w.Clear();

      m = (n + 1) / 2;
      xm = 0.5 * (b + a);
      xl = 0.5 * (b - a);

      for (i = 0; i < m; i++)
      {
        z = Math.Cos(Math.PI * (Convert.ToDouble(i) + 0.75) / (Convert.ToDouble(n) + 0.5));
        do{
          p1 = 1.0;
          p2 = 0.0;
          for (j = 0; j < n; j++)
          {
            p3 = p2;
            p2 = p1;
            p1 = ((2.0 * Convert.ToDouble(j) + 1) * z * p2 - Convert.ToDouble(j) * p3) /
                  Convert.ToDouble(j + 1);
          }
          pp = Convert.ToDouble(n) * (z * p1 - p2) / (z * z - 1);
          z1 = z;
          z = z1 - p1 / pp;
        } while(Math.Abs(z - z1) >= Constants.ZEROITER);

        x[i] = xm - xl * z;
        x[n - 1 - i] = xm + xl * z;
        w[i] = 2 * xl / ((1 - z * z) * pp * pp);
        w[n -1 - i] = w[i];
      }
    }
    void GauHerQ(Tensor x, Tensor w)   //-+ infinity
    {
      if (x.which_order != 1 || w.which_order != 1 || x.how_many_lines != w.how_many_lines)
      {
        throw new Exception("GaussQ: GauLegF: wrong tensor!!!");
      }

      int n = x.how_many_lines;
      int i, j, m;
      double p1, p2, p3, pp, z, z1;
      double PIM4;

      PIM4 = Math.Pow(Math.PI, -0.25);
      z = 0;

      m = (n + 1) / 2;
      for (i = 1; i <= m; i++)
      {
        if (i == 1)
        {
          z = Math.Sqrt(Convert.ToDouble(2 * n + 1)) - 1.85575 * Math.Pow(Convert.ToDouble(2 * n + 1), -0.16667);
        }
        else if (i == 2)
        {
          z -= 1.14 * Math.Pow(Convert.ToDouble(n), 0.426) / z;
        }
        else if (i == 3)
        {
          z = 1.86 * z - 0.86 * x[1];
        }
        else if (i == 4)
        {
          z = 1.91 * z - 0.91 * x[2];
        }
        else
        {
          z = 2.0 * z - x[i - 2];
        }
        do
        {
          p1 = PIM4;
          p2 = 0.0;
          for (j = 1; j <= n; j++)
          {
            p3 = p2;
            p2 = p1;
            p1 = z * Math.Sqrt(2 / Convert.ToDouble(j)) * p2 - Math.Sqrt((Convert.ToDouble(j - 1)) / j) * p3;
          }
          pp = Math.Sqrt(Convert.ToDouble(2 * n)) * p2;
          z1 = z;
          z = z1 - p1 / pp;
        } while (Math.Abs(z - z1) >= Constants.ZEROITER);

        x[i - 1] = z;
        x[n - i] = -z;
        w[i - 1] = 2.0 / (pp * pp);
        w[n - i] = w[i - 1];
      }
    }

  }
}